﻿using System.Collections.Immutable;
using System.Linq;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.Diagnostics;

namespace PersistenceCodeFix {
  [DiagnosticAnalyzer(LanguageNames.CSharp)]
  public sealed class ObsoleteStorableConstructorAnalyzer : DiagnosticAnalyzer {
    public const string DiagnosticId = "ObsoleteStorableConstructor";

    private static readonly LocalizableString Title = new LocalizableResourceString(
      nameof(Resources.ObsoleteStorableConstructorAnalyzerTitle), Resources.ResourceManager, typeof(Resources));

    private static readonly LocalizableString MessageFormat = new LocalizableResourceString(
      nameof(Resources.ObsoleteStorableConstructorAnalyzerMessageFormat), Resources.ResourceManager, typeof(Resources));

    private static readonly LocalizableString Description = new LocalizableResourceString(
      nameof(Resources.ObsoleteStorableConstructorAnalyzerDescription), Resources.ResourceManager, typeof(Resources));

    private const string Category = nameof(DiagnosticCategory.Persistence);

    private static readonly DiagnosticDescriptor Rule = new DiagnosticDescriptor(
      DiagnosticId, Title, MessageFormat, Category, DiagnosticSeverity.Error, isEnabledByDefault: true, description: Description);

    public override ImmutableArray<DiagnosticDescriptor> SupportedDiagnostics { get { return ImmutableArray.Create(Rule); } }

    public override void Initialize(AnalysisContext context) {
      context.RegisterSymbolAction(AnalyzeNamedType, SymbolKind.NamedType);
    }

    private static void AnalyzeNamedType(SymbolAnalysisContext context) {
      var namedTypeSymbol = (INamedTypeSymbol)context.Symbol;
      if (namedTypeSymbol.IsStatic) return;
      if (namedTypeSymbol.TypeKind == TypeKind.Delegate
          || namedTypeSymbol.TypeKind == TypeKind.Enum
          || namedTypeSymbol.TypeKind == TypeKind.Interface) return;

      foreach (var ctor in namedTypeSymbol.Constructors) {
        var attr = ctor.GetAttributes();
        if (attr.Any(x => x.AttributeClass.Name == "StorableConstructorAttribute")) {
          var @params = ctor.Parameters;
          if (@params.Length != 1 || @params[0].Type.SpecialType == SpecialType.System_Boolean) {
            var diagnostic = Diagnostic.Create(Rule, ctor.Locations[0], namedTypeSymbol.Name);
            context.ReportDiagnostic(diagnostic);
          }
        }
      }
    }
  }
}
